"""Utilities for describing capabilities.

Each known capability is assigned a base class for request and response.
The actual request and response types in a integration implementation
can either use the base classes directly or create subclasses, however,
those bases are enforced to be used.
"""

import inspect
import typing as t
from dataclasses import dataclass

from pydantic import BaseModel, ValidationError

from connector.generated import (
    ActivateAccountRequest,
    ActivateAccountResponse,
    AppInfoRequest,
    AppInfoResponse,
    AssignEntitlementRequest,
    AssignEntitlementResponse,
    AuthCredential,
    BasicCredential,
    CapabilitySchema,
    CreateAccountRequest,
    CreateAccountResponse,
    DeactivateAccountRequest,
    DeactivateAccountResponse,
    DeleteAccountRequest,
    DeleteAccountResponse,
    ErrorCode,
    FindEntitlementAssociationsRequest,
    FindEntitlementAssociationsResponse,
    GetAuthorizationUrlRequest,
    GetAuthorizationUrlResponse,
    GetLastActivityRequest,
    GetLastActivityResponse,
    HandleAuthorizationCallbackRequest,
    HandleAuthorizationCallbackResponse,
    HandleClientCredentialsRequest,
    HandleClientCredentialsResponse,
    JWTCredential,
    ListAccountsRequest,
    ListAccountsResponse,
    ListCustomAttributesSchemaRequest,
    ListCustomAttributesSchemaResponse,
    ListEntitlementsRequest,
    ListEntitlementsResponse,
    ListExpensesRequest,
    ListExpensesResponse,
    ListResourcesRequest,
    ListResourcesResponse,
    OAuth1Credential,
    OAuthClientCredential,
    OAuthCredential,
    Page,
    RefreshAccessTokenRequest,
    RefreshAccessTokenResponse,
    ServiceAccountCredential,
    StandardCapabilityName,
    TokenCredential,
    UnassignEntitlementRequest,
    UnassignEntitlementResponse,
    UpdateAccountRequest,
    UpdateAccountResponse,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
)
from connector.oai.errors import ConnectorError

BaseModelType = t.TypeVar("BaseModelType", bound=BaseModel)


class Request(t.Protocol):
    """
    A generic request to any capability.

    Useful as a type for calling request helpers on, e.g. get_token_auth, get_settings.

    Will match all AuthenticatedRequest capability inputs.
    """

    auth: t.Any
    """
    The connector.generated.AuthCredential attached to this request.
    """

    credentials: t.Any

    request: t.Any
    """
    The payload of this request. The type depends on the request type.
    """

    page: t.Any
    """
    Page data. May be None
    """

    include_raw_data: bool | None = None

    settings: t.Any
    """
    User-configured settings for the integration.
    """


class AuthRequest(t.Protocol):
    """
    A request being used as part of an authentication flow.

    These requests must have settings, but they don't have credentials.
    """

    request: t.Any
    """
    The payload of this request
    """

    page: t.Any
    """
    Page data. May be None
    """

    include_raw_data: bool | None = None

    settings: t.Any
    """
    User-configured settings for the integration.
    """


CredentialType = t.TypeVar(
    "CredentialType",
    OAuthCredential,
    OAuthClientCredential,
    BasicCredential,
    TokenCredential,
    JWTCredential,
)


def get_credential(
    request: Request, credential_id: str, credential_type: type[CredentialType]
) -> CredentialType:
    """
    Return the particular credential from the request.
    Similarly to get_settings, the credential is identified by the credential_id and the root credentials model.
    eg. OAuthCredential, OAuthClientCredential, BasicCredential, TokenCredential, JWTCredential
    """
    if request.credentials and isinstance(request.credentials, list):
        for credential in request.credentials:
            if credential.id == credential_id:
                # Find the credential from the request
                if credential.oauth and isinstance(credential.oauth, credential_type):
                    return credential.oauth
                elif credential.oauth_client_credentials and isinstance(
                    credential.oauth_client_credentials, credential_type
                ):
                    return credential.oauth_client_credentials
                elif credential.basic and isinstance(credential.basic, credential_type):
                    return credential.basic
                elif credential.token and isinstance(credential.token, credential_type):
                    return credential.token
                elif credential.jwt and isinstance(credential.jwt, credential_type):
                    return credential.jwt

                raise ConnectorError(
                    message=f"Credential '{credential_id}' found but is not of type {credential_type}.",
                    error_code=ErrorCode.BAD_REQUEST,
                )

    raise ConnectorError(
        message=f"Credential '{credential_id}' not provided in credentials.",
        error_code=ErrorCode.BAD_REQUEST,
    )


def get_oauth(request: Request) -> OAuthCredential | OAuthClientCredential:
    """
    Either get the (valid) request.auth as an OAuth credential, or throw an error.
    """
    if request.auth and request.auth.oauth and isinstance(request.auth.oauth, OAuthCredential):
        return request.auth.oauth
    if (
        request.auth
        and request.auth.oauth_client_credentials
        and isinstance(request.auth.oauth_client_credentials, OAuthClientCredential)
    ):
        return request.auth.oauth_client_credentials

    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_oauth_1(request: Request) -> OAuth1Credential:
    """
    Either get the (valid) request.auth as an OAuth1 credential, or throw an error.
    """
    if request.auth and request.auth.oauth1 and isinstance(request.auth.oauth1, OAuth1Credential):
        return request.auth.oauth1
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_basic_auth(request: Request) -> BasicCredential:
    """
    Either get the (valid) request.auth as a basic credential, or throw an error.
    """
    if request.auth and request.auth.basic and isinstance(request.auth.basic, BasicCredential):
        return request.auth.basic
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_token_auth(request: Request) -> TokenCredential:
    """
    Either get the (valid) request.auth as a token credential, or throw an error.
    """
    if request.auth and request.auth.token and isinstance(request.auth.token, TokenCredential):
        return request.auth.token
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_jwt_auth(request: Request) -> JWTCredential:
    """
    Either get the (valid) request.auth as a JWT credential, or throw an error.
    """
    if request.auth and request.auth.jwt and isinstance(request.auth.jwt, JWTCredential):
        return request.auth.jwt
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_service_account_auth(request: Request) -> ServiceAccountCredential:
    """
    Either get the (valid) request.auth as a service account credential, or throw an error.
    """
    if (
        request.auth
        and request.auth.service_account
        and isinstance(request.auth.service_account, ServiceAccountCredential)
    ):
        return request.auth.service_account
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_page(request: Request) -> Page:
    """
    Get request.page or a blank one if it isn't set.
    """
    if request.page:
        return request.page
    return Page()


SettingsType = t.TypeVar("SettingsType", bound=BaseModel)


def get_settings(
    request: Request | AuthRequest | AppInfoRequest, model: type[SettingsType]
) -> SettingsType:
    """
    Get a validated Settings type from request.settings.
    """
    try:
        return model.model_validate(request.settings or {})
    except ValidationError as err:
        raise ConnectorError(
            message="Invalid request settings", error_code=ErrorCode.BAD_REQUEST
        ) from err


def extra_data(extra: dict[str, t.Any]) -> dict[str, str]:
    """
    Generate a str:str dict out of arbitrary data for a response.raw_data.
    """
    ret: dict[str, str] = {}
    for key, value in extra.items():
        if value:
            ret[key] = str(value)
    return ret


T = t.TypeVar("T")


class Response(t.Protocol):
    response: t.Any
    raw_data: t.Any | None


_Request = t.TypeVar("_Request", bound=Request, contravariant=True)


class CapabilityCallableProto(t.Protocol, t.Generic[_Request]):
    def __call__(self, args: _Request) -> Response | t.Awaitable[Response]:
        ...

    __name__: str


class CustomRequest(BaseModel, t.Generic[BaseModelType]):
    """
    Generic Request type, extendable for customized capability inputs.

    Example:

    class MyAppCreateAccount(CreateAccount):
      birthday: str

    def create_account(args: CustomRequest[MyAppCreateAccount]) -> CreateAccountResponse:
       ...
    """

    auth: AuthCredential
    credentials: t.Any = None
    page: Page | None = None
    include_raw_data: bool | None = None
    settings: t.Any

    request: BaseModelType


class CustomResponse(BaseModel, t.Generic[BaseModelType]):
    page: Page | None = None
    raw_data: t.Any | None = None

    response: BaseModelType


class Empty(BaseModel):
    pass


def generate_capability_schema(
    impl: (CapabilityCallableProto[t.Any]),
    capability_description: str | None = None,
    full_schema: bool = False,
) -> CapabilitySchema:
    request_annotation, response_annotation = get_capability_annotations(impl)
    request_type = _request_payload_type(request_annotation)
    response_type = _response_payload_type(response_annotation)

    # Old behavior: use Empty for list types when full_schema is False
    if not full_schema:
        request_type = Empty if _is_list(request_type) else request_type
        response_type = Empty if _is_list(response_type) else response_type
        return CapabilitySchema(
            argument=request_type.model_json_schema(),
            output=response_type.model_json_schema(),
            description=capability_description,
        )

    # Get the full payload type
    request_type = _full_payload_type(request_annotation)
    response_type = _full_payload_type(response_annotation)

    # New behavior: handle full schema generation for both simple and list types
    base_request_type = request_type.__args__[0] if _is_list(request_type) else request_type
    base_response_type = response_type.__args__[0] if _is_list(response_type) else response_type

    base_request_schema = base_request_type.model_json_schema(
        ref_template="#/components/schemas/{model}"
    )
    base_response_schema = base_response_type.model_json_schema(
        ref_template="#/components/schemas/{model}"
    )

    request_schema = (
        {
            "type": "array",
            "items": base_request_schema,
            "title": f"Array of {getattr(base_request_type, '__name__', 'Items')}",
        }
        if _is_list(request_type)
        else base_request_schema
    )

    response_schema = (
        {
            "type": "array",
            "items": base_response_schema,
            "title": f"Array of {getattr(base_response_type, '__name__', 'Items')}",
        }
        if _is_list(response_type)
        else base_response_schema
    )

    return CapabilitySchema(
        argument=request_schema,
        output=response_schema,
        description=capability_description,
    )


def get_capability_annotations(
    impl: CapabilityCallableProto[t.Any],
) -> tuple[t.Any, t.Any]:
    """Extract argument and return type annotations."""
    annotations = inspect.get_annotations(impl)
    try:
        response_annotation = annotations["return"]
        request_annotation_name = (set(annotations.keys()) - {"return"}).pop()
    except KeyError:
        raise TypeError(
            f"The capability function {impl.__name__} must have both request and return annotations."
        ) from None

    request_annotation = annotations[request_annotation_name]
    return request_annotation, response_annotation


@dataclass
class CapabilitySignature:
    @dataclass
    class Payload:
        _type: type[BaseModel]
        envelope_type: type[BaseModel]
        is_list: bool
        json_schema: t.Any
        may_be_customized: bool = False

    input_payload: Payload
    output_payload: Payload


def _payload_type_data(
    *, envelope_type: t.Any, is_request: bool, may_be_customized=False
) -> CapabilitySignature.Payload:
    if is_request:
        payload = _request_payload_type(envelope_type)
    else:
        payload = _response_payload_type(envelope_type)
    return CapabilitySignature.Payload(
        envelope_type=envelope_type,
        _type=_pluck_generic_parameter(payload),
        is_list=_is_list(payload),
        json_schema=_pluck_generic_parameter(payload).model_json_schema(),
        may_be_customized=may_be_customized,
    )


def get_capability_signature(
    impl: CapabilityCallableProto[t.Any],
) -> CapabilitySignature:
    """Extract input and return types from a capability impl function."""
    annotations = inspect.get_annotations(impl)
    try:
        response_type = annotations["return"]
        request_annotation_name = (set(annotations.keys()) - {"return"}).pop()
    except KeyError:
        raise TypeError(
            f"The capability function {impl.__name__} must have both request and return annotations."
        ) from None

    return CapabilitySignature(
        input_payload=_payload_type_data(
            envelope_type=annotations[request_annotation_name], is_request=True
        ),
        output_payload=_payload_type_data(envelope_type=response_type, is_request=False),
    )


def validate_capability(
    capability_name: StandardCapabilityName,
    impl: (CapabilityCallableProto[t.Any]),
) -> None:
    """Make sure capability implementation is valid.

    Capability is marked as valid when:
        * is fully annotated, i.e., both argument and return value are
        type-hinted
        * type of accepted argument matches the expected one, i.e., is
        exactly the same class or a subclass
        * type of returned value matches the expected one, same
        mechanism as for argument
    """
    actual_signature = get_capability_signature(impl)
    expected_signature = _STANDARD_CAPABILITY_SIGNATURES[capability_name]
    for message_name, actual, expected in (
        ("input", actual_signature.input_payload, expected_signature.input_payload),
        ("output", actual_signature.output_payload, expected_signature.output_payload),
    ):
        if actual.is_list != expected.is_list:
            raise TypeError(
                f"{capability_name} {message_name} should {'not ' if not expected.is_list else ''}be a list"
            )
        if not expected.may_be_customized:
            if actual._type != expected._type:
                raise TypeError(
                    f"{capability_name} {message_name} can only be {expected._type.__name__}"
                )
            if actual.envelope_type != expected.envelope_type:
                raise TypeError(
                    f"{capability_name} {message_name} can only be {expected.envelope_type.__name__}"
                )
        else:
            actual_fields = actual.json_schema["properties"]
            expected_fields = expected.json_schema["properties"]
            for property in expected_fields:
                if property not in actual_fields:
                    raise TypeError(
                        f"{capability_name} {message_name} may not drop a field: {property}"
                    )
                if actual_fields[property].get("type", None) != expected_fields[property]["type"]:
                    raise TypeError(
                        f"{capability_name} {message_name} may not change the type of {property}"
                    )
            actual_required = set(actual.json_schema.get("required", []))
            expected_required = set(expected.json_schema.get("required", []))
            missing = expected_required - actual_required
            if missing:
                raise TypeError(
                    f"{capability_name} {message_name} doesn't keep as required: {', '.join(sorted(missing))}"
                )


def _full_payload_type(model: type[BaseModelType]) -> type[BaseModelType]:
    if not hasattr(model, "model_fields"):
        raise TypeError(f"Not a pydantic model: {model}")
    return model


def _request_payload_type(model: type[BaseModel]) -> t.Any:
    if not hasattr(model, "model_fields"):
        raise TypeError(f"Not a pydantic model: {model}")
    return model.model_fields["request"].annotation


def _response_payload_type(model: type[BaseModel]) -> t.Any:
    if not hasattr(model, "model_fields"):
        raise TypeError(f"Not a pydantic model: {model}")
    return model.model_fields["response"].annotation


def _pluck_generic_parameter(type_annotation: t.Any) -> t.Any:
    if hasattr(type_annotation, "__args__"):
        value_type = type_annotation.__args__[-1]
        return value_type
    return type_annotation


def _is_list(type_annotation: t.Any) -> bool:
    """This function is compatible with both list and typing.List
    (which is used in the generated models)"""
    if origin := getattr(type_annotation, "__origin__", None):
        return origin is list
    return type_annotation is list


def capability_requires_authentication(capability: CapabilityCallableProto[t.Any]) -> bool:
    """Check if the capability requires authentication on the request."""
    expected_request, _ = get_capability_annotations(capability)

    # TODO: Later on when "auth" is phased out, we can add the .is_required() check back here
    if "auth" in expected_request.model_fields or "credentials" in expected_request.model_fields:
        return True
    return False


_STANDARD_CAPABILITY_SIGNATURES: dict[StandardCapabilityName, CapabilitySignature] = {
    StandardCapabilityName.APP_INFO: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=AppInfoRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=AppInfoResponse, is_request=False),
    ),
    StandardCapabilityName.GET_AUTHORIZATION_URL: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=GetAuthorizationUrlRequest, is_request=True),
        output_payload=_payload_type_data(
            envelope_type=GetAuthorizationUrlResponse, is_request=False
        ),
    ),
    StandardCapabilityName.GET_LAST_ACTIVITY: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=GetLastActivityRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=GetLastActivityResponse, is_request=False),
    ),
    StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK: CapabilitySignature(
        input_payload=_payload_type_data(
            envelope_type=HandleAuthorizationCallbackRequest, is_request=True
        ),
        output_payload=_payload_type_data(
            envelope_type=HandleAuthorizationCallbackResponse, is_request=False
        ),
    ),
    StandardCapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST: CapabilitySignature(
        input_payload=_payload_type_data(
            envelope_type=HandleClientCredentialsRequest, is_request=True
        ),
        output_payload=_payload_type_data(
            envelope_type=HandleClientCredentialsResponse, is_request=False
        ),
    ),
    StandardCapabilityName.LIST_ACCOUNTS: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=ListAccountsRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=ListAccountsResponse, is_request=False),
    ),
    StandardCapabilityName.LIST_RESOURCES: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=ListResourcesRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=ListResourcesResponse, is_request=False),
    ),
    StandardCapabilityName.LIST_ENTITLEMENTS: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=ListEntitlementsRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=ListEntitlementsResponse, is_request=False),
    ),
    StandardCapabilityName.LIST_EXPENSES: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=ListExpensesRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=ListExpensesResponse, is_request=False),
    ),
    StandardCapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS: CapabilitySignature(
        input_payload=_payload_type_data(
            envelope_type=FindEntitlementAssociationsRequest, is_request=True
        ),
        output_payload=_payload_type_data(
            envelope_type=FindEntitlementAssociationsResponse, is_request=False
        ),
    ),
    StandardCapabilityName.LIST_CUSTOM_ATTRIBUTES_SCHEMA: CapabilitySignature(
        input_payload=_payload_type_data(
            envelope_type=ListCustomAttributesSchemaRequest, is_request=True
        ),
        output_payload=_payload_type_data(
            envelope_type=ListCustomAttributesSchemaResponse, is_request=False
        ),
    ),
    StandardCapabilityName.REFRESH_ACCESS_TOKEN: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=RefreshAccessTokenRequest, is_request=True),
        output_payload=_payload_type_data(
            envelope_type=RefreshAccessTokenResponse, is_request=False
        ),
    ),
    StandardCapabilityName.CREATE_ACCOUNT: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=CreateAccountRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=CreateAccountResponse, is_request=False),
    ),
    StandardCapabilityName.DELETE_ACCOUNT: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=DeleteAccountRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=DeleteAccountResponse, is_request=False),
    ),
    StandardCapabilityName.ACTIVATE_ACCOUNT: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=ActivateAccountRequest, is_request=True),
        output_payload=_payload_type_data(envelope_type=ActivateAccountResponse, is_request=False),
    ),
    StandardCapabilityName.DEACTIVATE_ACCOUNT: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=DeactivateAccountRequest, is_request=True),
        output_payload=_payload_type_data(
            envelope_type=DeactivateAccountResponse, is_request=False
        ),
    ),
    StandardCapabilityName.ASSIGN_ENTITLEMENT: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=AssignEntitlementRequest, is_request=True),
        output_payload=_payload_type_data(
            envelope_type=AssignEntitlementResponse, is_request=False
        ),
    ),
    StandardCapabilityName.UNASSIGN_ENTITLEMENT: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=UnassignEntitlementRequest, is_request=True),
        output_payload=_payload_type_data(
            envelope_type=UnassignEntitlementResponse, is_request=False
        ),
    ),
    StandardCapabilityName.UPDATE_ACCOUNT: CapabilitySignature(
        input_payload=_payload_type_data(
            envelope_type=UpdateAccountRequest, is_request=True, may_be_customized=True
        ),
        output_payload=_payload_type_data(
            envelope_type=UpdateAccountResponse,
            is_request=False,
            may_be_customized=True,
        ),
    ),
    StandardCapabilityName.VALIDATE_CREDENTIALS: CapabilitySignature(
        input_payload=_payload_type_data(envelope_type=ValidateCredentialsRequest, is_request=True),
        output_payload=_payload_type_data(
            envelope_type=ValidateCredentialsResponse, is_request=False
        ),
    ),
}



import base64
import hashlib
import hmac
import random
import string
import time
import urllib.parse
from typing import Any

import httpx

from connector.generated import OAuth1Credential


def generate_nonce(length: int = 16) -> str:
    return "".join(random.choice(string.ascii_letters + string.digits) for _ in range(length))


def get_timestamp() -> str:
    return str(int(time.time()))


def encode_parameter(value: str | bytes) -> str:
    if isinstance(value, str):
        value = value.encode("utf-8")
    return urllib.parse.quote(value, safe="").replace("%7E", "~")


def build_signature_base_string(http_method: str, base_url: str, params: dict[str, Any]) -> str:
    encoded_params = []
    for key, value in sorted(params.items()):
        encoded_params.append(f"{encode_parameter(key)}={encode_parameter(value)}")

    parameter_string = "&".join(encoded_params)

    signature_base = (
        encode_parameter(http_method.upper())
        + "&"
        + encode_parameter(base_url)
        + "&"
        + encode_parameter(parameter_string)
    )

    return signature_base


def build_signature(signature_base: str, consumer_secret: str, token_secret: str) -> str:
    key = encode_parameter(consumer_secret) + "&" + encode_parameter(token_secret)

    signature = hmac.new(
        key.encode("utf-8"), signature_base.encode("utf-8"), hashlib.sha256
    ).digest()

    return base64.b64encode(signature).decode("utf-8")


def generate_oauth1_header(
    consumer_key: str,
    consumer_secret: str,
    token_id: str,
    token_secret: str,
    url: str,
    realm: str | None = None,
    http_method: str = "GET",
    query_params: dict[str, Any] | None = None,
    oauth_body_hash: str | None = None,
):
    parsed_url = urllib.parse.urlparse(url)
    base_url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"

    url_params = {}
    if parsed_url.query:
        url_params = dict(urllib.parse.parse_qsl(parsed_url.query))

    combined_params = {}
    if query_params:
        combined_params.update(query_params)
    combined_params.update(url_params)

    oauth_params = {
        "oauth_consumer_key": consumer_key,
        "oauth_token": token_id,
        "oauth_nonce": generate_nonce(),
        "oauth_timestamp": get_timestamp(),
        "oauth_signature_method": "HMAC-SHA256",
        "oauth_version": "1.0",
    }

    if oauth_body_hash:
        oauth_params["oauth_body_hash"] = oauth_body_hash

    signature_params = {}
    signature_params.update(combined_params)
    signature_params.update(oauth_params)

    signature_base = build_signature_base_string(http_method, base_url, signature_params)
    signature = build_signature(signature_base, consumer_secret, token_secret)

    oauth_params["oauth_signature"] = signature

    auth_header = "OAuth "
    if realm:
        auth_header += f'realm="{realm}",'

    auth_header += ",".join(
        [f'{encode_parameter(k)}="{encode_parameter(v)}"' for k, v in oauth_params.items()]
    )

    return auth_header


def calculate_body_hash(body: bytes | dict[str, Any] | list[Any] | str | None):
    if body is None:
        body = ""
    elif isinstance(body, dict) or isinstance(body, list):
        import json

        body = json.dumps(body)
    elif not isinstance(body, str):
        body = str(body)

    if isinstance(body, str):
        body = body.encode("utf-8")

    hash_obj = hashlib.sha256(body)
    return base64.b64encode(hash_obj.digest()).decode("utf-8")


def use_oauth1_with_request(
    request: httpx.Request, credentials: OAuth1Credential, realm: str | None = None
) -> httpx.Request:
    consumer_key = credentials.consumer_key
    consumer_secret = credentials.consumer_secret
    token_id = credentials.token_id
    token_secret = credentials.token_secret

    http_method = str(request.method)
    url = str(request.url)

    oauth_body_hash = None
    if http_method.upper() in ["POST", "PUT", "PATCH"] and hasattr(request, "content"):
        oauth_body_hash = calculate_body_hash(request.content)

    query_params = {}
    if hasattr(request.url, "params") and request.url.params:
        query_params = dict(request.url.params)

    auth_header = generate_oauth1_header(
        consumer_key=consumer_key,
        consumer_secret=consumer_secret,
        token_id=token_id,
        token_secret=token_secret,
        realm=realm,
        url=url,
        http_method=http_method,
        query_params=query_params,
        oauth_body_hash=oauth_body_hash,
    )

    request.headers["Authorization"] = auth_header

    return request



"""Test cases for test_pagination_utils.

Those cases cover scenarios for checking that ``decode_pagination`` is
inverse of ``encode_pagination``.
"""

import typing as t

from tests.utils.definitions import Pagination

Case: t.TypeAlias = list[Pagination]


def case_empty() -> Case:
    """No paginations is allowed."""
    return []


def case_single_url_page() -> Case:
    """Pagination for one call with pagination."""
    paginations = [
        Pagination(endpoint="/users", page=1),
    ]
    return paginations


def case_two_url_page() -> Case:
    """Pagination for more calls with pagination.

    This is rare case when we know the next page will call for last page
    of one endpoint and thus we can append call for the first page of
    the next endpoint.
    """
    paginations = [
        Pagination(endpoint="/users", page=100),
        Pagination(endpoint="/license", page=1),
    ]
    return paginations



from typing import Any

import httpx


class BearerAuth(httpx.Auth):
    """Authorization schema for authorization via token in header with default Bearer prefix."""

    def __init__(
        self,
        token: str,
        token_prefix: str = "Bearer",
        auth_header: str = "Authorization",
        custom_headers: dict[str, Any] | None = None,
    ) -> None:
        if custom_headers is None:
            custom_headers = {"Content-Type": "application/json"}
        self._token = token
        self._token_prefix = token_prefix
        self._auth_header = auth_header
        self._custom_headers = custom_headers

    def auth_flow(self, request: httpx.Request):
        request.headers[
            self._auth_header
        ] = f"{self._token_prefix}{' ' if self._token_prefix else ''}{self._token}"
        request.headers.update(self._custom_headers)
        yield request



"""Test cases for test_pagination_utils.

Those cases cover scenarios for ``decode_pagination`` function.
"""


def case_decode_empty() -> str:
    """Empty token gives no pagination."""
    token = "90"
    return token



from datetime import datetime, timedelta, timezone

import jwt
from connector.generated import JWTClaims, JWTCredential, JWTHeaders
from connector.utils.jwt_utils import sign_jwt

# Create test credentials
test_claims = JWTClaims(
    iss="test-issuer",
    aud="test-audience",
    sub="test-subject",
    exp=int((datetime.now(timezone.utc) + timedelta(minutes=20)).timestamp()),
    nbf=int(datetime.now(timezone.utc).timestamp()),
    iat=int(datetime.now(timezone.utc).timestamp()),
    jti="test-jti",
    act="test-actor",
    scope=["test-scope"],
    client_id="test-client-id",
    may_act="test-may-act",
)
test_headers = JWTHeaders(
    kid="test-key-id",
    alg="HS256",
    jku="test-jku",
    jwk="test-jwk",
    typ="test-typ",
    x5u="test-x5u",
    x5c="test-x5c",
    x5t="test-x5t",
    cty="test-cty",
    crit=["test-crit"],
    **{"x5t#S256": "test-x5t-S256"},
)
credentials = JWTCredential(secret="test-secret", claims=test_claims, headers=test_headers)


def test_sign_jwt():
    # Sign JWT
    token = sign_jwt(credentials)

    # Decode and verify token
    decoded = jwt.decode(
        token,
        "test-secret",
        algorithms=["HS256"],
        audience="test-audience",
        options={"verify_signature": True},
    )

    # Verify claims
    assert decoded["iss"] == "test-issuer"
    assert decoded["aud"] == "test-audience"

    # Verify timestamps
    now = datetime.now(timezone.utc)

    # iat is the current time
    assert abs(decoded["iat"] - int(now.timestamp())) < 2  # Allow 2 sec difference

    # exp is the current time + 20 minutes
    assert abs(decoded["exp"] - int((now + timedelta(minutes=20)).timestamp())) < 2


def test_sign_jwt_custom_expiration():
    # Sign JWT with custom expiration
    token_custom = sign_jwt(credentials, expiration_minutes=30)

    # Verify custom expiration
    decoded_custom = jwt.decode(
        token_custom,
        "test-secret",
        algorithms=["HS256"],
        audience="test-audience",
        options={"verify_signature": True},
    )

    # Verify custom expiration
    now = datetime.now(timezone.utc)

    # exp is the current time + custom set expiration time 30 minutes
    assert abs(decoded_custom["exp"] - int((now + timedelta(minutes=30)).timestamp())) < 2



"""Cases for testing ``get_last_activity`` capability."""

import typing as t

import httpx
from connector.generated import (
    ActivityEventType,
    ErrorResponse,
    GetLastActivity,
    GetLastActivityRequest,
    GetLastActivityResponse,
    LastActivityData
)

from tests.common_mock_data import VALID_AUTH, SETTINGS
from connector.tests.type_definitions import MockedResponse, ResponseBodyMap

from connector.generated.models.standard_capability_name import StandardCapabilityName

Case: t.TypeAlias = tuple[
    StandardCapabilityName,
    GetLastActivityRequest,
    ResponseBodyMap,
    GetLastActivityResponse | ErrorResponse,
]


def case_one_user_200() -> Case:
    """Successful request."""
    args = GetLastActivityRequest(
        request=GetLastActivity(account_ids=["user_1_id"]),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )

    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = GetLastActivityResponse(
        response=[
            LastActivityData(
                account_id="user_1_id",
                event_type=ActivityEventType.LAST_LOGIN,
                happened_at="2024-07-01T20:43:09-10:00",
            )
        ],
        page=None,
    )
    return StandardCapabilityName.GET_LAST_ACTIVITY, args, response_body_map, expected_response



import json
import typing as t

import pytest_cases
from connector.generated import StandardCapabilityName
from connector.oai.integration import Integration


@pytest_cases.parametrize_with_cases(
    ["integration", "capability_name", "request_data", "expected_response"],
    cases=[
        "tests.oai.test_exception_handler_handled_cases",
    ],
)
async def test_handled_exception(
    integration: Integration,
    capability_name: StandardCapabilityName,
    request_data: str,
    expected_response: dict[str, t.Any],
) -> None:
    """Test various exception handlers."""
    response = await integration.dispatch(capability_name, request_data)
    assert json.loads(response) == expected_response



import typing as t

from graphql import DocumentNode, print_ast
from pytest_httpx import HTTPXMock

from connector.tests.type_definitions import ResponseBodyMap


def mock_graphql_request_body(
    query: DocumentNode, variables: dict[str, t.Any] | None = None
) -> dict[str, t.Any]:
    body: dict[str, t.Any] = {"query": print_ast(query)}

    if variables:
        body.update({"variables": variables})

    return body


def mock_requests(
    response_body_map: ResponseBodyMap, httpx_mock: HTTPXMock, *, host: str | None = None
):
    if not response_body_map:
        # Don't mock any requests, and use the default behavior of
        # httpx_mock which is HTTP 200, empty body
        return

    for method, requests in response_body_map.items():
        if isinstance(requests, dict):
            for request_line, response in requests.items():
                if isinstance(response.response_body, str):
                    httpx_mock.add_response(
                        method=method,
                        url=f"{host or ''}{request_line}",
                        content=response.response_body.encode(),
                        status_code=response.status_code,
                    )
                else:
                    httpx_mock.add_response(
                        method=method,
                        url=f"{host or ''}{request_line}",
                        json=response.response_body,
                        status_code=response.status_code,
                        headers=response.headers if response.headers else None,
                        match_json=response.request_json_body
                        if response.request_json_body
                        else None,
                    )
        elif isinstance(requests, list):
            for response in requests:
                if isinstance(response.response_body, str):
                    httpx_mock.add_response(
                        method=method,
                        url=host or "",
                        content=response.response_body.encode(),
                        status_code=response.status_code,
                    )
                else:
                    httpx_mock.add_response(
                        method=method,
                        url=host or "",
                        json=response.response_body,
                        status_code=response.status_code,
                        headers=response.headers if response.headers else None,
                        match_json=response.request_json_body
                        if response.request_json_body
                        else None,
                    )



import contextvars
import logging
from collections.abc import Generator
from contextlib import contextmanager

import httpx
from gql.transport.httpx import HTTPXAsyncTransport as GqlHTTPXAsyncTransport
from urllib3.util.url import Url, parse_url

from connector.generated import ErrorCode
from connector.oai.errors import ConnectorError

logger = logging.getLogger("integration-connectors.sdk")


LUMOS_PROXY_URL: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "LUMOS_PROXY_URL", default=None
)
LUMOS_PROXY_HEADERS: contextvars.ContextVar[httpx.Headers] = contextvars.ContextVar(
    "LUMOS_PROXY_HEADERS", default=httpx.Headers()
)


@contextmanager
def proxy_settings(
    *, proxy_url: str | None, proxy_headers: dict[str, str] | httpx.Headers | None = None
) -> Generator[None, None, None]:
    """
    Set context to run some httpx requests that are rewritten to a proxy.

    Example:

      from httpx_rewrite import AsyncClient, proxy_settings

      with proxy_settings(proxy_url="https://mitm.lumos.com/whatever/path", proxy_headers={"password":"swordfish"}):
          client = AsyncClient()
          await client.get("https://a-real-url-to-call/different/path")
    """
    LUMOS_PROXY_URL.set(proxy_url)
    LUMOS_PROXY_HEADERS.set(httpx.Headers(proxy_headers))
    yield


def get_proxy_url() -> str | None:
    """Return the current URL requests should proxy to. Set this with proxy_settings()"""
    return LUMOS_PROXY_URL.get()


def get_proxy_headers() -> httpx.Headers:
    """Return the current headers that should be attached to proxied requests. Set this with proxy_settings()"""
    return LUMOS_PROXY_HEADERS.get()


class AsyncClient(httpx.AsyncClient):
    """
    A slightly modified version of httpx.AsyncClient, that may rewrite requests to go to a proxy.

    Requests will be rewritten when run within the proxy_settings() context.

    Rewritten requests...
    1. will go to the proxy URL
    2. will have the original request line (including fragment) preserved in the header `X-Forward-To`
    3. will have their headers preserved, EXCEPT FOR...
       a. `Host` will be set to the proxy host
       b. Additional headers will be set, and overwrite any passed-in headers, from
         - the LUMOS_PROXY_HEADERS environment variable (as JSON)
         - the accumulated calls to this module's `add_proxy_headers` function
    """

    def will_rewrite_request(self) -> bool:
        return bool(get_proxy_url())

    async def send(self, request: httpx.Request, **kwargs) -> httpx.Response:
        proxy_host = get_proxy_url()
        if proxy_host:
            logger.warning("Proxying all requests to %s", proxy_host)
            override = parse_url(proxy_host)
        else:
            return await super().send(request, **kwargs)

        # Update request URL
        url = str(request.url)
        updated = Url(
            scheme=override.scheme,
            host=override.host,
            port=override.port,
            path=override.path,
            query=override.query,
        )

        # Rewrite headers
        new_headers = request.headers.copy()
        new_headers["X-Forward-To"] = url
        new_headers["Host"] = Url(host=override.host, port=override.port).url
        proxy_headers = get_proxy_headers()
        for header, value in proxy_headers.items():
            new_headers[header] = value

        # Get request body if applicable
        if request.method in ("POST", "PUT", "PATCH"):
            content = request.content or bytes([])
        else:
            content = None

        new_request = httpx.Request(
            method=request.method,
            url=updated.url,
            headers=new_headers,
            content=content,
        )

        # Log everything when using the proxy for debugging
        logger.debug("Sending request to proxy: %s", new_request.url)
        logger.debug("Request method: %s", new_request.method)
        response = await super().send(new_request, **kwargs)
        logger.debug("Request headers: %s", [(k, v) for k, v in response.request.headers.items()])
        logger.debug("Request body: %s", response.request.content)
        logger.debug("Response headers: %s", [(k, v) for k, v in response.headers.items()])
        truncated_response = response.text[:1000] + "..." if len(response.text) > 1000 else ""
        logger.debug("Response: %s", truncated_response)
        logger.debug("Response status code: %s", response.status_code)

        # Create a debug hint if the proxy/service is not working as expected
        proxy_status = int(response.headers.get("x-lumos-sip-status-code") or response.status_code)
        target_status = response.headers.get("x-lumos-target-hostname-status-code")
        if proxy_status != 200:
            hint = (
                "There is something wrong with the proxy itself or you are giving it a bad "
                f"arn value. Proxy returned code {proxy_status}; "
                "sort this out first before troubleshooting the target service."
            )
            hint_status = proxy_status
            hint_url = new_request.url

        elif target_status is not None and not str(target_status).startswith("2"):
            hint = (
                f"The target service returned code {target_status}, "
                "the proxy is working as expected."
            )
            hint_status = target_status
            hint_url = request.url

        if "hint" in locals():
            raise ConnectorError(
                error_code=ErrorCode.INVALID_RESPONSE,
                message=f"[{hint_status}][{hint_url}] {hint} Original response: {response.text}",
                app_error_code="proxy_mode_error_debug_hint",
            )

        return response


class HTTPXAsyncTransport(GqlHTTPXAsyncTransport):
    async def connect(self):
        # Call the superclass...
        await super().connect()

        # ... and then override its result
        self.client = AsyncClient(**self.kwargs)



import httpx
from connector.generated import OAuthCredential
from connector.oai.capability import StandardCapabilityName
from connector.oai.errors import HTTPHandler
from connector.oai.integration import DescriptionData, Integration

from {name}.__about__ import __version__
from {name}.enums import entitlement_types, resource_types
from {name}.settings import {pascal}Settings
from {name} import capabilities_read, capabilities_write


integration = Integration(
    app_id="{hyphenated_name}",
    version=__version__,
    auth=OAuthCredential,
    exception_handlers=[
        (httpx.HTTPStatusError, HTTPHandler, None),
    ],
    description_data=DescriptionData(
        logo_url="", user_friendly_name="{pascal}", description="", categories=[]
    ),
    settings_model={pascal}Settings,
    resource_types=resource_types,
    entitlement_types=entitlement_types,
)

integration.register_capabilities(
    {{
        # Read capabilities
        StandardCapabilityName.VALIDATE_CREDENTIALS: capabilities_read.validate_credentials,
        # StandardCapabilityName.LIST_ACCOUNTS: capabilities_read.list_accounts,
        # StandardCapabilityName.LIST_RESOURCES: capabilities_read.list_resources,
        # StandardCapabilityName.LIST_ENTITLEMENTS: capabilities_read.list_entitlements,
        # StandardCapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS: capabilities_read.find_entitlement_associations,
        # StandardCapabilityName.GET_LAST_ACTIVITY: capabilities_read.get_last_activity,
        # Write capabilities
        # StandardCapabilityName.ASSIGN_ENTITLEMENT: capabilities_write.assign_entitlement,
        # StandardCapabilityName.UNASSIGN_ENTITLEMENT: capabilities_write.unassign_entitlement,
        # StandardCapabilityName.CREATE_ACCOUNT: capabilities_write.create_account,
        # StandardCapabilityName.ACTIVATE_ACCOUNT: capabilities_write.activate_account,
        # StandardCapabilityName.DEACTIVATE_ACCOUNT: capabilities_write.deactivate_account,
        # StandardCapabilityName.DELETE_ACCOUNT: capabilities_write.delete_account,
    }}
)



from connector.generated import (
    AssignEntitlementRequest,
    AssignEntitlementResponse,
    ActivateAccountRequest,
    ActivateAccountResponse,
    CreateAccountRequest,
    CreateAccountResponse,
    DeactivateAccountRequest,
    DeactivateAccountResponse,
    DeleteAccountRequest,
    DeleteAccountResponse,
    UnassignEntitlementRequest,
    UnassignEntitlementResponse,
)


async def assign_entitlement(args: AssignEntitlementRequest) -> AssignEntitlementResponse:
    raise NotImplementedError


async def unassign_entitlement(
    args: UnassignEntitlementRequest,
) -> UnassignEntitlementResponse:
    raise NotImplementedError


async def create_account(
    args: CreateAccountRequest,
) -> CreateAccountResponse:
    raise NotImplementedError


async def delete_account(
    args: DeleteAccountRequest,
) -> DeleteAccountResponse:
    raise NotImplementedError


async def activate_account(
    args: ActivateAccountRequest,
) -> ActivateAccountResponse:
    raise NotImplementedError


async def deactivate_account(
    args: DeactivateAccountRequest,
) -> DeactivateAccountResponse:
    raise NotImplementedError



"""Cases for testing ``validate_credentials`` capability."""

import typing as t

from connector.generated import (
    ErrorResponse,
    ValidateCredentials,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)

from tests.common_mock_data import SETTINGS, VALID_AUTH
from connector.tests.type_definitions import ResponseBodyMap

from connector.generated.models.standard_capability_name import StandardCapabilityName

Case: t.TypeAlias = tuple[
    StandardCapabilityName,
    ValidateCredentialsRequest,
    ResponseBodyMap,
    ValidateCredentialsResponse | ErrorResponse,
]


def case_validate_credentials_200() -> Case:
    """Successful request."""
    args = ValidateCredentialsRequest(
        request=ValidateCredentials(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map: ResponseBodyMap = {{}}
    expected_response = ValidateCredentialsResponse(
        response=ValidatedCredentials(valid=True, unique_tenant_id="REPLACE_WITH_UNIQUE_TENANT_ID"),
    )
    return StandardCapabilityName.VALIDATE_CREDENTIALS, args, response_body_map, expected_response



import typing as t
from abc import abstractmethod
from pathlib import Path

from gql import Client
from gql.client import AsyncClientSession
from gql.dsl import DSLSchema
from graphql import GraphQLSchema, build_schema
from typing_extensions import Self

from connector.generated import ErrorCode
from connector.httpx_rewrite import AsyncClient
from connector.oai.capability import Request
from connector.oai.errors import ConnectorError


class BaseIntegrationClient:
    @classmethod
    @abstractmethod
    def prepare_client_args(cls, args: Request) -> dict[str, t.Any]:
        pass

    @classmethod
    def build_client(cls, args: Request) -> AsyncClient:
        return AsyncClient(**cls.prepare_client_args(args))

    def __init__(self, args: Request) -> None:
        self._http_client = self.build_client(args)

    async def __aenter__(self):
        await self._http_client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._http_client.__aexit__()
        if exc_type is not None:
            raise exc_val


class BaseGraphQLSession(AsyncClientSession):
    def __init__(self, args: Request):
        super().__init__(client=self.build_client(args))

    async def __aenter__(self) -> Self:
        await self.client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.client.__aexit__(exc_type=exc_type, exc=exc, tb=tb)

        if exc_type is not None:
            raise exc

    @classmethod
    @abstractmethod
    def prepare_client_args(cls, args: Request) -> dict[str, t.Any]:
        pass

    @classmethod
    def build_client(cls, args: Request) -> Client:
        return Client(**cls.prepare_client_args(args))

    @classmethod
    def load_schema(cls, schema_file_path: str | Path) -> GraphQLSchema:
        """Load the GraphQL schema from a file."""
        with open(schema_file_path) as f:
            return build_schema(f.read())

    @property
    def schema(self) -> DSLSchema:
        if self.client.schema is None:
            raise ConnectorError(
                message="Failed to fetch schema",
                error_code=ErrorCode.API_ERROR,
            )

        return DSLSchema(self.client.schema)



from connector.generated import (
    OAuthCredential,
    StandardCapabilityName,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from connector.oai.integration import DescriptionData, Integration
from connector.oai.modules.oauth_module_types import OAuthSettings


async def test_scopes_for_capabilities():
    """Test if scopes are set for each implemented capability."""
    app_id = "test"
    integration = Integration(
        app_id=app_id,
        version="0.1.0",
        auth=OAuthCredential,
        oauth_settings=OAuthSettings(
            authorization_url="https://test.com/authorize",
            token_url="https://test.com/token",
            scopes={
                StandardCapabilityName.VALIDATE_CREDENTIALS: "test",
            },
        ),
        description_data=DescriptionData(user_friendly_name="test_scopes.py", categories=[]),
        exception_handlers=[],
        handle_errors=True,
    )

    @integration.register_capability(StandardCapabilityName.VALIDATE_CREDENTIALS)
    async def validate_credentials(args: ValidateCredentialsRequest) -> ValidateCredentialsResponse:
        return ValidateCredentialsResponse(
            response=ValidatedCredentials(
                valid=True,
                unique_tenant_id="test",
            ),
        )

    for capability in integration.capabilities:
        assert integration.oauth_settings and integration.oauth_settings.scopes is not None
        if capability not in [
            "app_info",
            "get_authorization_url",
            "handle_authorization_callback",
            "refresh_access_token",
        ]:
            assert StandardCapabilityName(capability) in integration.oauth_settings.scopes



"""Utility class to manage connector implementation.

:py:class:`Integration` provides a single point to register Integration
capabilities.

By instantiating the :py:class:`Integration` you simply create a basic
integration without any real implementation. To actually implement any
capability, you have to define (async) function outside the class and
register them to the integration instance by decorating the
implementation with ``@integration.register_capability(name)``.

Capability function has to:
    * accept only one argument
    * return scalar response

The :py:class:`Integration` is as much type-hinted as possible and also
does several checks to ensure that the implementation "is correct".
Incorrect implementation should raise an error during application start
(fail fast).

What is checked (at application start):
    * capability name is known (defined in ``StandardCapabilityName`` enum)
    * the types of accepted argument and returned value matches the
    capability interface
"""

import inspect
import json
import logging
import re
import typing as t
from dataclasses import asdict, dataclass

from pydantic import BaseModel, ValidationError

from connector.generated import (
    AppCategory,
    AuthCredential,
    AuthModel,
    BasicCredential,
    CapabilitySchema,
    CredentialConfig,
    EntitlementType,
    Error,
    ErrorCode,
    ErrorResponse,
    Info,
    InfoResponse,
    JWTCredential,
    OAuth1Credential,
    OAuthClientCredential,
    OAuthCredential,
    ResourceType,
    ServiceAccountCredential,
    StandardCapabilityName,
    TokenCredential,
)
from connector.oai.capability import (
    CapabilityCallableProto,
    capability_requires_authentication,
    generate_capability_schema,
    get_capability_annotations,
    validate_capability,
)
from connector.oai.errors import ErrorMap, handle_exception
from connector.oai.modules.base_module import BaseIntegrationModule
from connector.oai.modules.info_module import InfoModule
from connector.oai.modules.oauth_module import OAuthModule
from connector.oai.modules.oauth_module_types import OAuthConfig, OAuthSettings

AuthSetting: t.TypeAlias = (
    type[OAuthCredential]
    | type[OAuthClientCredential]
    | type[OAuth1Credential]
    | type[BasicCredential]
    | type[TokenCredential]
    | type[JWTCredential]
    | type[ServiceAccountCredential]
)


AUTH_TYPE_MAP = {
    AuthModel.OAUTH: OAuthCredential,
    AuthModel.OAUTH_CLIENT_CREDENTIALS: OAuthClientCredential,
    AuthModel.BASIC: BasicCredential,
    AuthModel.TOKEN: TokenCredential,
    AuthModel.JWT: JWTCredential,
    AuthModel.SERVICE_ACCOUNT: ServiceAccountCredential,
}


logger = logging.getLogger("integration-connectors.sdk")


class IntegrationError(Exception):
    """Base class for exceptions raised by Integration."""


class DuplicateCapabilityError(IntegrationError):
    """Raised when registering the same capability repeatedly."""

    def __init__(self, capability_name: str) -> None:
        super().__init__(f"{capability_name} already registered")


class ReservedCapabilityNameError(IntegrationError):
    """
    Raised when registering a custom capability, using a name reserved for standard capabilities
    """

    def __init__(self, capability_name: str) -> None:
        super().__init__(
            f"Cannot register {capability_name} as a custom capability. This name is reserved for "
            f"the standard {capability_name} capability."
        )


class InvalidCapabilityNameError(IntegrationError):
    """
    Raised when registering a custom capability, using a name that violates our naming conventions
    """

    def __init__(self, capability_name: str, reason: str | None) -> None:
        msg = f"Cannot register {capability_name} as a custom capability."
        if reason is None:
            msg += (
                " Name violates our naming conventions. Please make sure it is snake cased,"
                " does not contain numbers or special characters."
            )
        else:
            msg += f" {reason}"
        super().__init__(msg)


class InvalidAppIdError(IntegrationError):
    """Raised when app_id is not valid.

    Most probably, empty or containing only whitespaces.
    """


@dataclass
class DescriptionData:
    user_friendly_name: str
    categories: list[AppCategory]
    description: str | None = None
    logo_url: str | None = None


@dataclass
class CapabilityMetadata:
    display_name: str | None = None
    description: str | None = None


class EmptySettings(BaseModel):
    pass


def _is_snake_case(text: str) -> bool:
    snake_case_pattern = re.compile(r"^[a-z0-9]+(?:_[a-z0-9]+)*$")
    return bool(snake_case_pattern.match(text))


def _validate_capability_name(capability_name: str) -> None:
    is_snake_case = _is_snake_case(capability_name)
    if not is_snake_case:
        raise InvalidCapabilityNameError(
            capability_name, reason="Capability names must use snake casing."
        )

    if not capability_name.replace("_", "").isalpha():
        raise InvalidCapabilityNameError(
            capability_name,
            reason="Capability names must only contain alphabetic characters and underscores",
        )


class Integration:
    app_id: str
    settings_model: type[BaseModel]

    def __init__(
        self,
        *,
        app_id: str,
        version: str,
        exception_handlers: ErrorMap,
        description_data: DescriptionData,
        auth: AuthSetting | None = None,
        credentials: t.Sequence[CredentialConfig | OAuthConfig] | None = None,
        handle_errors: bool = True,
        resource_types: list[ResourceType] | None = None,
        entitlement_types: list[EntitlementType] | None = None,
        settings_model: type[BaseModel] | None = None,
        oauth_settings: OAuthSettings | None = None,
    ):
        self.app_id = app_id.strip()
        self.version = version
        self.auth = auth
        self.credentials = credentials or []
        self.exception_handlers = exception_handlers
        self.handle_errors = handle_errors
        self.description_data = description_data
        self.resource_types = resource_types or []
        self.entitlement_types = entitlement_types or []
        self.settings_model = settings_model or EmptySettings
        self.oauth_settings = oauth_settings

        if len(self.app_id) == 0:
            raise InvalidAppIdError

        self.capabilities: dict[str, CapabilityCallableProto[t.Any]] = {}
        self.capability_metadata: dict[str, CapabilityMetadata] = {}
        self.modules: list[BaseIntegrationModule] = []

        # Add OAuthModule if credentials contain OAuthConfig
        if self.credentials and any(isinstance(cred, OAuthConfig) for cred in self.credentials):
            oauth_credentials = [
                cred
                for cred in self.credentials
                if isinstance(cred, OAuthConfig) and cred.oauth_settings is not None
            ]
            self.add_module(OAuthModule(oauth_credentials))
        else:
            # Add OAuthModule if oauth_settings is provided (backwards compatibility, until phasing 'auth' out)
            if self.oauth_settings is not None:
                self.add_module(OAuthModule([self.oauth_settings]))

        # Attach the info module (app_info capability)
        self.add_module(InfoModule())

    def add_module(self, module: BaseIntegrationModule):
        self.modules.append(module)
        module.register(self)

    def register_capability(
        self,
        name: StandardCapabilityName,
        display_name: str | None = None,
        description: str | None = None,
    ) -> t.Callable[
        [CapabilityCallableProto[t.Any]],
        CapabilityCallableProto[t.Any],
    ]:
        """Add implementation of specified capability.

        This function is expected to be used as a decorator for a
        capability implementation.

        Display name is an optional text to display for the integration capability,
        in place of the default behavior with is to use the capability name.

        Description is an optional 1-2 sentence description of any unusual behaviors
        of the capability, rendered for the end user.

        Raises
        ------
        RuntimeError:
            When capability is registered more that once.
        """
        if name.value in self.capabilities:
            raise DuplicateCapabilityError(name.value)

        def decorator(
            func: CapabilityCallableProto[t.Any],
        ) -> CapabilityCallableProto[t.Any]:
            validate_capability(name, func)
            self.capabilities[name.value] = func
            self.capability_metadata[name.value] = CapabilityMetadata(
                display_name=display_name, description=description
            )
            return func

        return decorator

    def register_custom_capability(
        self,
        name: str,
        display_name: str | None = None,
        description: str | None = None,
    ) -> t.Callable[
        [CapabilityCallableProto[t.Any]],
        CapabilityCallableProto[t.Any],
    ]:
        """Add implementation of specified custom capability.

        This function is expected to be used as a decorator for a
        custom capability implementation.

        Custom capabilities differ from standard capabilities, because:
        - They can have any capability name (as long as it meets naming standards)
        - There is more flexibility around input and output types

        Display name is an optional text to display for the integration capability,
        in place of the default behavior with is to use the capability name.

        Description is an optional 1-2 sentence description of any unusual behaviors
        of the capability, rendered for the end user.

        Raises
        ------
        RuntimeError:
            When capability is registered more that once.
        """
        if name in {
            standard_capability_name.value for standard_capability_name in StandardCapabilityName
        }:
            raise ReservedCapabilityNameError(name)

        _validate_capability_name(name)

        if name in self.capabilities:
            raise DuplicateCapabilityError(name)

        def decorator(
            func: CapabilityCallableProto[t.Any],
        ) -> CapabilityCallableProto[t.Any]:
            self.capabilities[name] = func
            self.capability_metadata[name] = CapabilityMetadata(
                display_name=display_name,
                description=description,
            )
            return func

        return decorator

    def register_capabilities(
        self,
        capabilities: dict[
            StandardCapabilityName,
            CapabilityCallableProto[t.Any]
            | tuple[CapabilityCallableProto[t.Any], CapabilityMetadata],
        ],
    ):
        """
        Register a dictionary of capabilities.

        This is a convenience method to register multiple capabilities at once.
        You can pass a tuple of (capability, metadata) to register a capability, along with any
        configuration specific to that capability, such as display_name, description, etc

        eg.
        ```
        integration.register_capabilities({
            StandardCapabilityName.VALIDATE_CREDENTIALS: (
                validate_credentials, CapabilityMetadata(description="Test")
            ),
            StandardCapabilityName.LIST_ACCOUNTS: list_accounts,
        })
        ```
        """
        for name, capability in capabilities.items():
            if isinstance(capability, tuple):
                func, metadata = capability
                self.register_capability(name, **asdict(metadata))(func)
            else:
                self.register_capability(name)(capability)

    async def dispatch(self, name: str, request_string: str) -> str:
        """Call implemented capability, returning the result.

        Raises
        ------
        NotImplementedError:
            When capability is not implemented (or registered)
        """
        try:
            capability = self.capabilities[name]
        except KeyError:
            if self.handle_errors:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message=f"Capability '{name}' is not implemented.",
                        error_code=ErrorCode.NOT_IMPLEMENTED,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
            raise NotImplementedError from None

        # JSON Input
        json_object = json.loads(request_string)
        if not json_object:
            return ErrorResponse(
                is_error=True,
                error=Error(
                    message="Invalid request, expected JSON input.",
                    error_code=ErrorCode.BAD_REQUEST,
                    app_id=self.app_id,
                ),
            ).model_dump_json()

        # Is this a call to the app_info capability?
        is_app_info_call = name == StandardCapabilityName.APP_INFO.value

        # Credentials parameter validation
        validated_credentials = False
        if "credentials" in json_object and json_object["credentials"] is not None:
            """
            TypeSpec does not support this level of union oneOf types,
            hence we do the validation here.
            At some point in the future, we should just remove 'auth', but for now we are
            backwards compatible and it is being kept.
            """
            if len(json_object["credentials"]) == 0 and self._is_using_credentials():
                if not is_app_info_call:
                    return ErrorResponse(
                        is_error=True,
                        error=Error(
                            message="Missing credentials in request",
                            error_code=ErrorCode.BAD_REQUEST,
                            app_id=self.app_id,
                        ),
                    ).model_dump_json()

            if "auth" in json_object and json_object["auth"] is not None:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Cannot pass credentials and auth in the same request, if you've meant to make this connector multi-auth compatible, please remove the auth field.",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()

            # Validate the credentials
            input_credentials: list[AuthCredential] = []
            try:
                for index, credential in enumerate(json_object["credentials"]):
                    auth_credential = AuthCredential.model_validate(credential)
                    if not auth_credential.id:
                        return ErrorResponse(
                            is_error=True,
                            error=Error(
                                message=f"Missing ID in credential at index {index}",
                                error_code=ErrorCode.BAD_REQUEST,
                                app_id=self.app_id,
                            ),
                        ).model_dump_json()
                    else:
                        # Validate that we are receiving expected credentials
                        if auth_credential.id not in [c.id for c in self.credentials]:
                            return ErrorResponse(
                                is_error=True,
                                error=Error(
                                    message=f"Credential with id '{auth_credential.id}' not expected",
                                    error_code=ErrorCode.BAD_REQUEST,
                                    app_id=self.app_id,
                                ),
                            ).model_dump_json()

                    input_credentials.append(auth_credential)
            except ValidationError:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Malformed credentials in request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()

            for credential in input_credentials:
                # Grab its connector counterpart
                connector_credential = next(
                    (cred for cred in self.credentials if cred.id == credential.id), None
                )

                # If the credential is not expected, return an error
                if connector_credential is None:
                    return ErrorResponse(
                        is_error=True,
                        error=Error(
                            message=f"Credential with id '{credential.id}' not expected",
                            error_code=ErrorCode.BAD_REQUEST,
                            app_id=self.app_id,
                        ),
                    ).model_dump_json()

                # If the credential is optional, skip model validation
                if connector_credential.optional:
                    logger.info(
                        f"[{self.app_id}][INFO] Credential with id '{credential.id}' is optional, skipping model validation."
                    )
                    continue

                # Validate the credential type
                missing = "Nothing"
                if connector_credential.type == AuthModel.BASIC:
                    if not credential.basic:
                        missing = "basic"
                elif connector_credential.type == AuthModel.OAUTH:
                    if not credential.oauth:
                        missing = "oauth"
                elif connector_credential.type == AuthModel.OAUTH_CLIENT_CREDENTIALS:
                    if not credential.oauth_client_credentials:
                        missing = "oauth_client_credentials"
                elif connector_credential.type == AuthModel.TOKEN:
                    if not credential.token:
                        missing = "token"
                elif connector_credential.type == AuthModel.JWT:
                    if not credential.jwt:
                        missing = "jwt"
                elif connector_credential.type == AuthModel.SERVICE_ACCOUNT:
                    if not credential.service_account:
                        missing = "service_account"
                if missing != "Nothing":
                    return ErrorResponse(
                        is_error=True,
                        error=Error(
                            message=f"Missing '{missing}' credential in request",
                            error_code=ErrorCode.BAD_REQUEST,
                            app_id=self.app_id,
                        ),
                    ).model_dump_json()

            validated_credentials = True

        # TODO: Later on when "auth" is phased out, we can turn this on
        # elif capability_has_authenticated_request(capability):
        #    return ErrorResponse(
        #        is_error=True,
        #        error=Error(
        #            message="Missing credentials in request",
        #            error_code=ErrorCode.BAD_REQUEST,
        #            app_id=self.app_id,
        #        ),
        #    ).model_dump_json()

        # Auth parameter validation
        if "auth" in json_object and json_object["auth"] is not None:
            if "credentials" in json_object and json_object["credentials"] is not None:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Cannot pass credentials and auth in the same request, if you've meant to make this connector multi-auth compatible, please remove the auth field.",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()

            try:
                auth_credential = AuthCredential.model_validate(json_object["auth"])
            except ValidationError:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Malformed auth in request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
            missing = "Nothing"
            if self.auth == BasicCredential:
                if not auth_credential.basic:
                    missing = "basic"
            elif self.auth == OAuthCredential:
                if not auth_credential.oauth:
                    missing = "oauth"
            if missing != "Nothing":
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message=f"Missing '{missing}' auth in request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
        elif capability_requires_authentication(capability) and not validated_credentials:
            # Since this is not migrated yet, we just flat out skip is_app_info_call
            # In the future we can re-instate the is_required() checks inside this validation
            # and make this work for is_app_info_call automatically (its not required for is_app_info_call)
            if not is_app_info_call:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Missing auth/credentials in request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()

        # Capability annotation validation
        request_annotation, _ = get_capability_annotations(capability)
        try:
            request = request_annotation.model_validate(json_object)
        except ValidationError as e:
            return ErrorResponse(
                is_error=True,
                error=Error(
                    message=f"Invalid request - {repr(e.errors())}",
                    error_code=ErrorCode.BAD_REQUEST,
                    app_id=self.app_id,
                ),
            ).model_dump_json()

        # Parse and validate the connector settings before dispatching
        if self.settings_model == EmptySettings:
            pass
        elif request.settings is None and self.handle_errors:
            if not is_app_info_call:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message=f"No settings passed on request: {self.app_id} requires {self.settings_model.__name__}",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
        else:
            try:
                self.settings_model.model_validate(request.settings)
            except ValidationError:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Invalid settings passed on request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()

        try:
            if inspect.iscoroutinefunction(capability):
                response = await capability(request)
            else:
                response = t.cast(
                    BaseModel,
                    capability(request),
                )

            return response.model_dump_json()
        except Exception as e:
            return handle_exception(
                self.exception_handlers, e, capability, self.app_id
            ).model_dump_json()

    def info(self) -> InfoResponse:
        """Provide information about implemented capabilities.

        Json schema describing implemented capabilities and their
        interface is returned. The authentication schema is also
        included.
        """
        logger.debug("Generating info response")
        capability_names = sorted(self.capabilities.keys())
        capability_schema: dict[str, CapabilitySchema] = {}
        for capability_name in capability_names:
            command_types = generate_capability_schema(self.capabilities[capability_name])

            capability_metadata: CapabilityMetadata | None = None
            if capability_name in self.capability_metadata:
                capability_metadata = self.capability_metadata[capability_name]

            display_name: str | None = capability_name.replace("_", " ").title()
            if (
                capability_metadata is not None
                and capability_metadata.display_name is not None
                and isinstance(capability_metadata.display_name, str)
            ):
                display_name = capability_metadata.display_name

            description = None
            if (
                capability_metadata is not None
                and capability_metadata.description is not None
                and isinstance(capability_metadata.description, str)
            ):
                description = capability_metadata.description

            capability_schema[capability_name] = CapabilitySchema(
                argument=command_types.argument,
                output=command_types.output,
                display_name=display_name,
                description=description,
            )

        oauth_scopes = None
        if self.oauth_settings:
            if not callable(self.oauth_settings.scopes) and isinstance(
                self.oauth_settings.scopes, dict
            ):
                oauth_scopes = {
                    capability: scope
                    for capability, scope in self.oauth_settings.scopes.items()
                    if scope is not None
                }

        return InfoResponse(
            response=Info(
                app_id=self.app_id,
                version=self.version,
                capabilities=capability_names,
                capability_schema=capability_schema,
                logo_url=self.description_data.logo_url,
                user_friendly_name=self.description_data.user_friendly_name,
                description=self.description_data.description,
                categories=self.description_data.categories,
                resource_types=self.resource_types,
                entitlement_types=self.entitlement_types,
                request_settings_schema=self.get_model_extended_json_schema(self.settings_model),
                authentication_schema=self.get_model_extended_json_schema(self.auth)
                if self.auth
                else {},
                credentials_schema=self.get_credentials_json_schema() if self.credentials else [],
                oauth_scopes=oauth_scopes,
            )
        )

    def _is_using_credentials(self) -> bool:
        return len(self.credentials) > 0

    def get_model_extended_json_schema(self, model: type[BaseModel]) -> dict[str, t.Any]:
        json_schema = model.model_json_schema()
        field_order = list(model.model_fields.keys())
        json_schema["field_order"] = field_order
        return json_schema

    def get_credentials_json_schema(self) -> list[dict[str, t.Any]]:
        credentials = []

        if self.credentials:
            for credential in self.credentials:
                credential_schema = {}

                if credential.input_model:
                    # Custom input model (override)
                    credential_schema = self.get_model_extended_json_schema(
                        t.cast(type[BaseModel], credential.input_model)
                    )
                else:
                    # Standard input model
                    credential_schema = self.get_model_extended_json_schema(
                        t.cast(type[BaseModel], AUTH_TYPE_MAP[credential.type])
                    )

                credential_schema["id"] = credential.id
                credential_schema["description"] = credential.description

                """
                Very basic extra oauth_settings info
                TODO:
                - The issue here still is that this does not support multiple oauth credentials (not like we need that right now)
                - This should later be handled exclusively in the new info module and its response
                """
                scopes = None
                if self.oauth_settings:
                    oauth = self.oauth_settings
                    if oauth.scopes is not None:
                        if not callable(oauth.scopes) and isinstance(oauth.scopes, dict):
                            scopes = {
                                capability: scope
                                for capability, scope in oauth.scopes.items()
                                if scope is not None
                            }

                    credential_schema["oauth_settings"] = {
                        "oauth_type": oauth.flow_type,
                        "scopes": scopes,
                        "pkce_enabled": oauth.pkce,
                    }

                credentials.append(credential_schema)

        return credentials



import json

import pydantic
from connector.generated import (
    ErrorResponse,
)
from connector.oai.capability import StandardCapabilityName
from connector.tests.gather_cases import gather_cases
from connector.tests.mock_httpx import mock_requests
from connector.tests.type_definitions import ResponseBodyMap
from {name}.constants import BASE_URL
from {name}.integration import integration
from pytest_cases import parametrize_with_cases
from pytest_httpx import HTTPXMock

# If you want to test specific capability than replace gather_cases with list of cases.
INTEGRATION_CASES = gather_cases(integration)


@parametrize_with_cases(
    ["capability_name", "args", "response_body_map", "expected_response"],
    cases=INTEGRATION_CASES,
)
async def test_all_capabilities(
    capability_name: StandardCapabilityName,
    args: pydantic.BaseModel,
    response_body_map: ResponseBodyMap,
    expected_response: pydantic.BaseModel | ErrorResponse,
    httpx_mock: HTTPXMock,
) -> None:
    """Test all registered capabilities.
    Cases MUST:
        - be within any directory in <connector_name>/tests or tests themselves
        - file name in this format: "test_{{capability_name}}_cases.py"
    Good example of file location:
        - {name}/tests/read_capabilities/test_list_accounts_cases.py
    """
    mock_requests(response_body_map, httpx_mock, host=BASE_URL)
    response = await integration.dispatch(capability_name, args.model_dump_json())

    assert json.loads(response) == expected_response.model_dump()



import os
from enum import Enum
from pathlib import Path


class LogLevel(Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


LOG_LEVEL_ENV_VAR = "LUMOS_LOG_LEVEL"
LOG_DIRECTORY_ENV_VAR = "LUMOS_LOG_DIRECTORY"
LOG_TO_STDOUT_ENV_VAR = "LUMOS_LOG_TO_STDOUT"


class Config:
    def __init__(self) -> None:
        self.log_level: LogLevel = LogLevel.ERROR
        self.log_directory: Path | None = None
        self.log_to_stdout: bool = False

        self.set_log_level()
        self.set_log_directory()
        self.set_log_to_stdout()

    def set_log_level(self) -> None:
        log_level_string = os.environ.get(LOG_LEVEL_ENV_VAR, "ERROR")
        try:
            self.log_level = LogLevel(log_level_string)
        except ValueError as e:
            raise ValueError(
                f"Invalid log level set by environment variable {LOG_LEVEL_ENV_VAR}: {log_level_string}"
            ) from e

    def set_log_directory(self) -> None:
        log_directory_string = os.environ.get(LOG_DIRECTORY_ENV_VAR, None)
        if log_directory_string is not None:
            self.log_directory = Path(log_directory_string)

    def set_log_to_stdout(self) -> None:
        log_to_stdout_string = os.environ.get(LOG_TO_STDOUT_ENV_VAR, "False")
        self.log_to_stdout = log_to_stdout_string.lower() == "true"


config = Config()
